@@ -8,9 +8,12 @@ import (
 	"path/filepath"
 	"time"
 
-	"github.com/hashicorp/go-multierror"
+	multierror "github.com/hashicorp/go-multierror"
+	"github.com/hashicorp/go-version"
 	"github.com/hashicorp/hcl"
 	"github.com/hashicorp/hcl/hcl/ast"
+	"github.com/hashicorp/nomad/helper"
+	"github.com/hashicorp/nomad/helper/tlsutil"
 	"github.com/hashicorp/nomad/nomad/structs/config"
 	"github.com/mitchellh/mapstructure"
 )
@@ -91,11 +94,15 @@ func parseConfig(result *Config, list *ast.ObjectList) error {
 		"syslog_facility",
 		"disable_update_check",
 		"disable_anonymous_signature",
-		"atlas",
 		"consul",
+		"vault",
+		"tls",
 		"http_api_response_headers",
+		"acl",
+		"sentinel",
+		"autopilot",
 	}
-	if err := checkHCLKeys(list, valid); err != nil {
+	if err := helper.CheckHCLKeys(list, valid); err != nil {
 		return multierror.Prefix(err, "config:")
 	}
 
@@ -111,9 +118,13 @@ func parseConfig(result *Config, list *ast.ObjectList) error {
 	delete(m, "client")
 	delete(m, "server")
 	delete(m, "telemetry")
-	delete(m, "atlas")
 	delete(m, "consul")
+	delete(m, "vault")
+	delete(m, "tls")
 	delete(m, "http_api_response_headers")
+	delete(m, "acl")
+	delete(m, "sentinel")
+	delete(m, "autopilot")
 
 	// Decode the rest
 	if err := mapstructure.WeakDecode(m, result); err != nil {
@@ -155,6 +166,13 @@ func parseConfig(result *Config, list *ast.ObjectList) error {
 		}
 	}
 
+	// Parse ACL config
+	if o := list.Filter("acl"); len(o.Items) > 0 {
+		if err := parseACL(&result.ACL, o); err != nil {
+			return multierror.Prefix(err, "acl ->")
+		}
+	}
+
 	// Parse telemetry config
 	if o := list.Filter("telemetry"); len(o.Items) > 0 {
 		if err := parseTelemetry(&result.Telemetry, o); err != nil {
@@ -162,13 +180,6 @@ func parseConfig(result *Config, list *ast.ObjectList) error {
 		}
 	}
 
-	// Parse atlas config
-	if o := list.Filter("atlas"); len(o.Items) > 0 {
-		if err := parseAtlas(&result.Atlas, o); err != nil {
-			return multierror.Prefix(err, "atlas ->")
-		}
-	}
-
 	// Parse the consul config
 	if o := list.Filter("consul"); len(o.Items) > 0 {
 		if err := parseConsulConfig(&result.Consul, o); err != nil {
@@ -176,6 +187,34 @@ func parseConfig(result *Config, list *ast.ObjectList) error {
 		}
 	}
 
+	// Parse the vault config
+	if o := list.Filter("vault"); len(o.Items) > 0 {
+		if err := parseVaultConfig(&result.Vault, o); err != nil {
+			return multierror.Prefix(err, "vault ->")
+		}
+	}
+
+	// Parse the TLS config
+	if o := list.Filter("tls"); len(o.Items) > 0 {
+		if err := parseTLSConfig(&result.TLSConfig, o); err != nil {
+			return multierror.Prefix(err, "tls ->")
+		}
+	}
+
+	// Parse Sentinel config
+	if o := list.Filter("sentinel"); len(o.Items) > 0 {
+		if err := parseSentinel(&result.Sentinel, o); err != nil {
+			return multierror.Prefix(err, "sentinel->")
+		}
+	}
+
+	// Parse Autopilot config
+	if o := list.Filter("autopilot"); len(o.Items) > 0 {
+		if err := parseAutopilot(&result.Autopilot, o); err != nil {
+			return multierror.Prefix(err, "autopilot->")
+		}
+	}
+
 	// Parse out http_api_response_headers fields. These are in HCL as a list so
 	// we need to iterate over them and merge them.
 	if headersO := list.Filter("http_api_response_headers"); len(headersO.Items) > 0 {
@@ -208,7 +247,7 @@ func parsePorts(result **Ports, list *ast.ObjectList) error {
 		"rpc",
 		"serf",
 	}
-	if err := checkHCLKeys(listVal, valid); err != nil {
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
 		return err
 	}
 
@@ -240,7 +279,7 @@ func parseAddresses(result **Addresses, list *ast.ObjectList) error {
 		"rpc",
 		"serf",
 	}
-	if err := checkHCLKeys(listVal, valid); err != nil {
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
 		return err
 	}
 
@@ -272,7 +311,7 @@ func parseAdvertise(result **AdvertiseAddrs, list *ast.ObjectList) error {
 		"rpc",
 		"serf",
 	}
-	if err := checkHCLKeys(listVal, valid); err != nil {
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
 		return err
 	}
 
@@ -315,15 +354,25 @@ func parseClient(result **ClientConfig, list *ast.ObjectList) error {
 		"node_class",
 		"options",
 		"meta",
+		"chroot_env",
 		"network_interface",
 		"network_speed",
+		"memory_total_mb",
+		"cpu_total_compute",
 		"max_kill_timeout",
 		"client_max_port",
 		"client_min_port",
 		"reserved",
 		"stats",
+		"gc_interval",
+		"gc_disk_usage_threshold",
+		"gc_inode_usage_threshold",
+		"gc_parallel_destroys",
+		"gc_max_allocs",
+		"no_host_uuid",
+		"server_join",
 	}
-	if err := checkHCLKeys(listVal, valid); err != nil {
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
 		return err
 	}
 
@@ -334,11 +383,21 @@ func parseClient(result **ClientConfig, list *ast.ObjectList) error {
 
 	delete(m, "options")
 	delete(m, "meta")
+	delete(m, "chroot_env")
 	delete(m, "reserved")
 	delete(m, "stats")
+	delete(m, "server_join")
 
 	var config ClientConfig
-	if err := mapstructure.WeakDecode(m, &config); err != nil {
+	dec, err := mapstructure.NewDecoder(&mapstructure.DecoderConfig{
+		DecodeHook:       mapstructure.StringToTimeDurationHookFunc(),
+		WeaklyTypedInput: true,
+		Result:           &config,
+	})
+	if err != nil {
+		return err
+	}
+	if err := dec.Decode(m); err != nil {
 		return err
 	}
 
@@ -370,6 +429,20 @@ func parseClient(result **ClientConfig, list *ast.ObjectList) error {
 		}
 	}
 
+	// Parse out chroot_env fields. These are in HCL as a list so we need to
+	// iterate over them and merge them.
+	if chrootEnvO := listVal.Filter("chroot_env"); len(chrootEnvO.Items) > 0 {
+		for _, o := range chrootEnvO.Elem().Items {
+			var m map[string]interface{}
+			if err := hcl.DecodeObject(&m, o.Val); err != nil {
+				return err
+			}
+			if err := mapstructure.WeakDecode(m, &config.ChrootEnv); err != nil {
+				return err
+			}
+		}
+	}
+
 	// Parse reserved config
 	if o := listVal.Filter("reserved"); len(o.Items) > 0 {
 		if err := parseReserved(&config.Reserved, o); err != nil {
@@ -377,6 +450,13 @@ func parseClient(result **ClientConfig, list *ast.ObjectList) error {
 		}
 	}
 
+	// Parse ServerJoin config
+	if o := listVal.Filter("server_join"); len(o.Items) > 0 {
+		if err := parseServerJoin(&config.ServerJoin, o); err != nil {
+			return multierror.Prefix(err, "server_join->")
+		}
+	}
+
 	*result = &config
 	return nil
 }
@@ -406,7 +486,7 @@ func parseReserved(result **Resources, list *ast.ObjectList) error {
 		"iops",
 		"reserved_ports",
 	}
-	if err := checkHCLKeys(listVal, valid); err != nil {
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
 		return err
 	}
 
@@ -450,17 +530,32 @@ func parseServer(result **ServerConfig, list *ast.ObjectList) error {
 		"bootstrap_expect",
 		"data_dir",
 		"protocol_version",
+		"raft_protocol",
 		"num_schedulers",
 		"enabled_schedulers",
 		"node_gc_threshold",
+		"eval_gc_threshold",
+		"job_gc_threshold",
+		"deployment_gc_threshold",
 		"heartbeat_grace",
+		"min_heartbeat_ttl",
+		"max_heartbeats_per_second",
+		"rejoin_after_leave",
+		"encrypt",
+		"authoritative_region",
+		"non_voting_server",
+		"redundancy_zone",
+		"upgrade_version",
+
+		"server_join",
+
+		// For backwards compatibility
 		"start_join",
 		"retry_join",
 		"retry_max",
 		"retry_interval",
-		"rejoin_after_leave",
 	}
-	if err := checkHCLKeys(listVal, valid); err != nil {
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
 		return err
 	}
 
@@ -469,8 +564,123 @@ func parseServer(result **ServerConfig, list *ast.ObjectList) error {
 		return err
 	}
 
+	delete(m, "server_join")
+
 	var config ServerConfig
-	if err := mapstructure.WeakDecode(m, &config); err != nil {
+	dec, err := mapstructure.NewDecoder(&mapstructure.DecoderConfig{
+		DecodeHook:       mapstructure.StringToTimeDurationHookFunc(),
+		WeaklyTypedInput: true,
+		Result:           &config,
+	})
+	if err != nil {
+		return err
+	}
+	if err := dec.Decode(m); err != nil {
+		return err
+	}
+
+	if config.UpgradeVersion != "" {
+		if _, err := version.NewVersion(config.UpgradeVersion); err != nil {
+			return fmt.Errorf("error parsing upgrade_version: %v", err)
+		}
+	}
+
+	// Parse ServerJoin config
+	if o := listVal.Filter("server_join"); len(o.Items) > 0 {
+		if err := parseServerJoin(&config.ServerJoin, o); err != nil {
+			return multierror.Prefix(err, "server_join->")
+		}
+	}
+
+	*result = &config
+	return nil
+}
+
+func parseServerJoin(result **ServerJoin, list *ast.ObjectList) error {
+	list = list.Elem()
+	if len(list.Items) > 1 {
+		return fmt.Errorf("only one 'server_join' block allowed")
+	}
+
+	// Get our object
+	listVal := list.Items[0].Val
+
+	// Check for invalid keys
+	valid := []string{
+		"start_join",
+		"retry_join",
+		"retry_max",
+		"retry_interval",
+	}
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
+		return err
+	}
+
+	var m map[string]interface{}
+	if err := hcl.DecodeObject(&m, listVal); err != nil {
+		return err
+	}
+
+	var serverJoinInfo ServerJoin
+	dec, err := mapstructure.NewDecoder(&mapstructure.DecoderConfig{
+		DecodeHook:       mapstructure.StringToTimeDurationHookFunc(),
+		WeaklyTypedInput: true,
+		Result:           &serverJoinInfo,
+	})
+	if err != nil {
+		return err
+	}
+	if err := dec.Decode(m); err != nil {
+		return err
+	}
+
+	*result = &serverJoinInfo
+	return nil
+}
+
+func parseACL(result **ACLConfig, list *ast.ObjectList) error {
+	list = list.Elem()
+	if len(list.Items) > 1 {
+		return fmt.Errorf("only one 'acl' block allowed")
+	}
+
+	// Get our server object
+	obj := list.Items[0]
+
+	// Value should be an object
+	var listVal *ast.ObjectList
+	if ot, ok := obj.Val.(*ast.ObjectType); ok {
+		listVal = ot.List
+	} else {
+		return fmt.Errorf("acl value: should be an object")
+	}
+
+	// Check for invalid keys
+	valid := []string{
+		"enabled",
+		"token_ttl",
+		"policy_ttl",
+		"replication_token",
+	}
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
+		return err
+	}
+
+	var m map[string]interface{}
+	if err := hcl.DecodeObject(&m, listVal); err != nil {
+		return err
+	}
+
+	var config ACLConfig
+	dec, err := mapstructure.NewDecoder(&mapstructure.DecoderConfig{
+		DecodeHook:       mapstructure.StringToTimeDurationHookFunc(),
+		WeaklyTypedInput: true,
+		Result:           &config,
+	})
+	if err != nil {
+		return err
+	}
+	if err := dec.Decode(m); err != nil {
 		return err
 	}
 
@@ -492,9 +702,30 @@ func parseTelemetry(result **Telemetry, list *ast.ObjectList) error {
 		"statsite_address",
 		"statsd_address",
 		"disable_hostname",
+		"use_node_name",
 		"collection_interval",
+		"publish_allocation_metrics",
+		"publish_node_metrics",
+		"datadog_address",
+		"datadog_tags",
+		"prometheus_metrics",
+		"circonus_api_token",
+		"circonus_api_app",
+		"circonus_api_url",
+		"circonus_submission_interval",
+		"circonus_submission_url",
+		"circonus_check_id",
+		"circonus_check_force_metric_activation",
+		"circonus_check_instance_id",
+		"circonus_check_search_tag",
+		"circonus_check_display_name",
+		"circonus_check_tags",
+		"circonus_broker_id",
+		"circonus_broker_select_tag",
+		"disable_tagged_metrics",
+		"backwards_compatible_metrics",
 	}
-	if err := checkHCLKeys(listVal, valid); err != nil {
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
 		return err
 	}
 
@@ -518,39 +749,6 @@ func parseTelemetry(result **Telemetry, list *ast.ObjectList) error {
 	return nil
 }
 
-func parseAtlas(result **AtlasConfig, list *ast.ObjectList) error {
-	list = list.Elem()
-	if len(list.Items) > 1 {
-		return fmt.Errorf("only one 'atlas' block allowed")
-	}
-
-	// Get our atlas object
-	listVal := list.Items[0].Val
-
-	// Check for invalid keys
-	valid := []string{
-		"infrastructure",
-		"token",
-		"join",
-		"endpoint",
-	}
-	if err := checkHCLKeys(listVal, valid); err != nil {
-		return err
-	}
-
-	var m map[string]interface{}
-	if err := hcl.DecodeObject(&m, listVal); err != nil {
-		return err
-	}
-
-	var atlas AtlasConfig
-	if err := mapstructure.WeakDecode(m, &atlas); err != nil {
-		return err
-	}
-	*result = &atlas
-	return nil
-}
-
 func parseConsulConfig(result **config.ConsulConfig, list *ast.ObjectList) error {
 	list = list.Elem()
 	if len(list.Items) > 1 {
@@ -567,18 +765,23 @@ func parseConsulConfig(result **config.ConsulConfig, list *ast.ObjectList) error
 		"auto_advertise",
 		"ca_file",
 		"cert_file",
+		"checks_use_advertise",
 		"client_auto_join",
 		"client_service_name",
+		"client_http_check_name",
 		"key_file",
 		"server_auto_join",
 		"server_service_name",
+		"server_http_check_name",
+		"server_serf_check_name",
+		"server_rpc_check_name",
 		"ssl",
 		"timeout",
 		"token",
 		"verify_ssl",
 	}
 
-	if err := checkHCLKeys(listVal, valid); err != nil {
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
 		return err
 	}
 
@@ -604,30 +807,182 @@ func parseConsulConfig(result **config.ConsulConfig, list *ast.ObjectList) error
 	return nil
 }
 
-func checkHCLKeys(node ast.Node, valid []string) error {
-	var list *ast.ObjectList
-	switch n := node.(type) {
-	case *ast.ObjectList:
-		list = n
-	case *ast.ObjectType:
-		list = n.List
-	default:
-		return fmt.Errorf("cannot check HCL keys of type %T", n)
+func parseTLSConfig(result **config.TLSConfig, list *ast.ObjectList) error {
+	list = list.Elem()
+	if len(list.Items) > 1 {
+		return fmt.Errorf("only one 'tls' block allowed")
 	}
 
-	validMap := make(map[string]struct{}, len(valid))
-	for _, v := range valid {
-		validMap[v] = struct{}{}
+	// Get the TLS object
+	listVal := list.Items[0].Val
+
+	valid := []string{
+		"http",
+		"rpc",
+		"verify_server_hostname",
+		"rpc_upgrade_mode",
+		"ca_file",
+		"cert_file",
+		"key_file",
+		"verify_https_client",
+		"tls_cipher_suites",
+		"tls_min_version",
+		"tls_prefer_server_cipher_suites",
 	}
 
-	var result error
-	for _, item := range list.Items {
-		key := item.Keys[0].Token.Value().(string)
-		if _, ok := validMap[key]; !ok {
-			result = multierror.Append(result, fmt.Errorf(
-				"invalid key: %s", key))
-		}
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
+		return err
 	}
 
-	return result
+	var m map[string]interface{}
+	if err := hcl.DecodeObject(&m, listVal); err != nil {
+		return err
+	}
+
+	var tlsConfig config.TLSConfig
+	if err := mapstructure.WeakDecode(m, &tlsConfig); err != nil {
+		return err
+	}
+
+	if _, err := tlsutil.ParseCiphers(tlsConfig.TLSCipherSuites); err != nil {
+		return err
+	}
+
+	if _, err := tlsutil.ParseMinVersion(tlsConfig.TLSMinVersion); err != nil {
+		return err
+	}
+
+	*result = &tlsConfig
+	return nil
+}
+
+func parseVaultConfig(result **config.VaultConfig, list *ast.ObjectList) error {
+	list = list.Elem()
+	if len(list.Items) > 1 {
+		return fmt.Errorf("only one 'vault' block allowed")
+	}
+
+	// Get our Vault object
+	listVal := list.Items[0].Val
+
+	// Check for invalid keys
+	valid := []string{
+		"address",
+		"allow_unauthenticated",
+		"enabled",
+		"task_token_ttl",
+		"ca_file",
+		"ca_path",
+		"cert_file",
+		"create_from_role",
+		"key_file",
+		"tls_server_name",
+		"tls_skip_verify",
+		"token",
+	}
+
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
+		return err
+	}
+
+	var m map[string]interface{}
+	if err := hcl.DecodeObject(&m, listVal); err != nil {
+		return err
+	}
+
+	vaultConfig := config.DefaultVaultConfig()
+	dec, err := mapstructure.NewDecoder(&mapstructure.DecoderConfig{
+		DecodeHook:       mapstructure.StringToTimeDurationHookFunc(),
+		WeaklyTypedInput: true,
+		Result:           &vaultConfig,
+	})
+	if err != nil {
+		return err
+	}
+	if err := dec.Decode(m); err != nil {
+		return err
+	}
+
+	*result = vaultConfig
+	return nil
+}
+
+func parseSentinel(result **config.SentinelConfig, list *ast.ObjectList) error {
+	list = list.Elem()
+	if len(list.Items) > 1 {
+		return fmt.Errorf("only one 'sentinel' block allowed")
+	}
+
+	// Get our sentinel object
+	obj := list.Items[0]
+
+	// Value should be an object
+	var listVal *ast.ObjectList
+	if ot, ok := obj.Val.(*ast.ObjectType); ok {
+		listVal = ot.List
+	} else {
+		return fmt.Errorf("sentinel value: should be an object")
+	}
+
+	// Check for invalid keys
+	valid := []string{
+		"import",
+	}
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
+		return err
+	}
+
+	var config config.SentinelConfig
+	if err := hcl.DecodeObject(&config, listVal); err != nil {
+		return err
+	}
+
+	*result = &config
+	return nil
+}
+
+func parseAutopilot(result **config.AutopilotConfig, list *ast.ObjectList) error {
+	list = list.Elem()
+	if len(list.Items) > 1 {
+		return fmt.Errorf("only one 'autopilot' block allowed")
+	}
+
+	// Get our Autopilot object
+	listVal := list.Items[0].Val
+
+	// Check for invalid keys
+	valid := []string{
+		"cleanup_dead_servers",
+		"server_stabilization_time",
+		"last_contact_threshold",
+		"max_trailing_logs",
+		"enable_redundancy_zones",
+		"disable_upgrade_migration",
+		"enable_custom_upgrades",
+	}
+
+	if err := helper.CheckHCLKeys(listVal, valid); err != nil {
+		return err
+	}
+
+	var m map[string]interface{}
+	if err := hcl.DecodeObject(&m, listVal); err != nil {
+		return err
+	}
+
+	autopilotConfig := config.DefaultAutopilotConfig()
+	dec, err := mapstructure.NewDecoder(&mapstructure.DecoderConfig{
+		DecodeHook:       mapstructure.StringToTimeDurationHookFunc(),
+		WeaklyTypedInput: true,
+		Result:           &autopilotConfig,
+	})
+	if err != nil {
+		return err
+	}
+	if err := dec.Decode(m); err != nil {
+		return err
+	}
+
+	*result = autopilotConfig
+	return nil
 }
